# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import os
import sys
import shutil
from logging import INFO, getLogger
from logging.handlers import QueueHandler
from pathlib import Path
from queue import Empty, SimpleQueue
from threading import Lock
from typing import Optional, Sequence

from .._session_user import SessionUser
from .._subprocess import LoggingSubprocess
from .._logging import LoggerAdapter

_internal_logger_lock = Lock()
_internal_logger = getLogger("openjd_sessions_runner_base_internal_logger")
_internal_logger_adapter = LoggerAdapter(_internal_logger, extra=dict())
_internal_logger.setLevel(INFO)
_internal_logger.propagate = False
_internal_logger_queue: SimpleQueue = SimpleQueue()
_internal_logger.addHandler(QueueHandler(_internal_logger_queue))


def locate_windows_executable(
    args: Sequence[str],
    user: Optional[SessionUser],
    os_env_vars: Optional[dict[str, Optional[str]]],
    working_dir: str,
) -> Sequence[str]:
    cmd_path = Path(args[0])
    if cmd_path.is_absolute():
        # If it's an absolute path (e.g. C:\Foo\Bar.exe or C:\Foo\Bar) then we just return
        # and leave it up to the OS to resolve the executable's extention.

        # TODO: Do we actually still want to do the find as a check to see if the command exists &
        # is executable? This would catch stuff like 'c:\Foo\test.ps1' as a command (which fails)
        return args

    return_args = list(args)
    if user is None:
        return_args[0] = _locate_for_same_user(cmd_path, os_env_vars, working_dir)
    else:
        return_args[0] = _locate_for_other_user(cmd_path, os_env_vars, working_dir, user)
    return return_args


def _get_path_var_for_shutil_which(
    os_env_vars: Optional[dict[str, Optional[str]]], working_dir: str
):
    path_var: Optional[str] = None
    if os_env_vars:
        env_var_keys = {k.lower(): k for k in os_env_vars}
        path_var = os_env_vars.get(env_var_keys["path"]) if "path" in env_var_keys else None
    if path_var is None:
        path_var = os.environ.get("PATH", "")
    path_var = "%s;%s" % (working_dir, path_var)
    return path_var


def _locate_for_same_user(
    command: Path, os_env_vars: Optional[dict[str, Optional[str]]], working_dir: str
) -> str:
    path_var: Optional[str] = _get_path_var_for_shutil_which(os_env_vars, working_dir)
    exe = str(shutil.which(str(command), path=path_var))
    if not exe:
        raise RuntimeError("Could not find executable file: %s" % command)
    return exe


def _locate_for_other_user(
    command: Path,
    os_env_vars: Optional[dict[str, Optional[str]]],
    working_dir: str,
    user: SessionUser,
) -> str:  # pragma: nocover
    # Running as a potentially different user, so it's possible that
    # this process doesn't have read access to the executable file's location.
    # Thus, we need to rely on running a subprocess as the user to be able
    # to find the executable.

    if len(command.parts) > 1:
        # Windows cannot find executables by relative location
        # i.e. where "dir\test.bat"
        #
        # Even if that worked, we'd have to prepend the relative part of the command
        # to the path and then search for only the command.name. But, we don't generally
        # have the user's PATH env var value.
        #
        # So, for both of those reasons we just return the command and let the action fail out
        # naturally.
        return str(command)

    # Prevent issues that might arise by having multiple Actions trying to start up
    # concurrently -- grab a lock.
    with _internal_logger_lock:
        # Drain the message queue to ensure nothing remains from previous runs.
        try:
            while True:
                _internal_logger_queue.get(block=False)
        except Empty:
            pass  # Will happen when the queue is fully empty

        # When running in a service context, we want to call the non-service Python binary
        sys_executable = sys.executable.lower().replace("pythonservice.exe", "python.exe")

        # In the subprocess code, we avoid exit code 1 as that is returned if a Python exception is thrown.
        exit_code_success = 2
        exit_code_could_not_find_exe = 3

        path_var: Optional[str] = _get_path_var_for_shutil_which(os_env_vars, working_dir)
        process = LoggingSubprocess(
            logger=_internal_logger_adapter,
            args=[
                sys_executable,
                "-c",
                # Command injection here is possible, but it's irrelevant. The command is running
                # as the given user. No need for an attacker to be fancy here, they could just run
                # the desired attack command directly in the job template.
                "import shutil, sys, pathlib\n"
                + f"cmd = shutil.which({str(command)!r}, path={path_var!r})\n"
                + "if cmd:\n"
                + "  print(str(pathlib.Path(cmd).absolute()))\n"
                + f"  sys.exit({exit_code_success})\n"
                + f"sys.exit({exit_code_could_not_find_exe})\n",
            ],
            user=user,
            os_env_vars=os_env_vars,
            working_dir=str(working_dir),
        )
        process.run()  # blocking call
        exit_code = process.exit_code

        # We're seeing random errors when trying to run an Action's command immediately after this
        # outside of Session 0; theory is that maybe this has something to do with running two
        # CreateProcessWithLogonW calls back-to-back with little time inbetween. So, explicitly
        # delete the process object to try to force some cleanup of handles that maybe help the
        # profile get unloaded. (this seems like it might be doing the trick)
        # Error:
        #  [WinError 1018] Illegal operation attempted on a registry key that has been marked for deletion
        del process

        if exit_code == exit_code_could_not_find_exe:
            raise RuntimeError(f"Could not find executable file: {command}")

        # Parse the output
        try:
            while True:
                record = _internal_logger_queue.get(block=False)
                message = record.getMessage()
                if "Output:" in message:
                    break

            if exit_code == exit_code_success:
                exe_record = _internal_logger_queue.get(block=False)
                # The line of output with the result of 'shutil.which' is the location of the command
                return exe_record.getMessage()
        except Empty:
            raise RuntimeError(
                f"Could not run Python as user {user.user} to find executable {command} in PATH.\n"
                + f"The host configuration must allow users to run {sys_executable}."
            )

        # Collect the error output from the subprocess
        error_messages = []
        try:
            while True:
                record = _internal_logger_queue.get(block=False)
                error_messages.append(record.getMessage())
        except Empty:
            pass

        # Something went wrong in launching sys_executable.
        # Because this scenario may be difficult to diagnose, we include more context.
        raise RuntimeError(
            f"Could not run Python as user {user.user} to find executable {command} in PATH.\n"
            + f"The host configuration must allow users to run {sys_executable}.\n\nError output:\n"
            + "\n".join(error_messages)
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""This module contains code for interacting with Linux capabilities. The module uses the ctypes
module from the Python standard library to wrap the libcap library.

See https://man7.org/linux/man-pages/man7/capabilities.7.html for details on this Linux kernel
feature.
"""

import ctypes
import os
import sys
from contextlib import contextmanager
from ctypes.util import find_library
from enum import Enum
from functools import cache
from typing import Any, Generator, Optional, Tuple, TYPE_CHECKING


from .._logging import LOG


# Capability sets
CAP_EFFECTIVE = 0
CAP_PERMITTED = 1
CAP_INHERITABLE = 2

# Capability bit numbers
CAP_KILL = 5

# Values for cap_flag_value_t arguments
CAP_CLEAR = 0
CAP_SET = 1

cap_flag_t = ctypes.c_int
cap_flag_value_t = ctypes.c_int
cap_value_t = ctypes.c_int


class CapabilitySetType(Enum):
    INHERITABLE = CAP_INHERITABLE
    PERMITTED = CAP_PERMITTED
    EFFECTIVE = CAP_EFFECTIVE


class UserCapHeader(ctypes.Structure):
    _fields_ = [
        ("version", ctypes.c_uint32),
        ("pid", ctypes.c_int),
    ]


class UserCapData(ctypes.Structure):
    _fields_ = [
        ("effective", ctypes.c_uint32),
        ("permitted", ctypes.c_uint32),
        ("inheritable", ctypes.c_uint32),
    ]


class Cap(ctypes.Structure):
    _fields_ = [
        ("head", UserCapHeader),
        ("data", UserCapData),
    ]


if TYPE_CHECKING:
    cap_t = ctypes._Pointer[Cap]
    cap_flag_value_ptr = ctypes._Pointer[cap_flag_value_t]
    cap_value_ptr = ctypes._Pointer[cap_value_t]
    ssize_ptr_t = ctypes._Pointer[ctypes.c_ssize_t]
else:
    cap_t = ctypes.POINTER(Cap)
    cap_flag_value_ptr = ctypes.POINTER(cap_flag_value_t)
    cap_value_ptr = ctypes.POINTER(cap_value_t)
    ssize_ptr_t = ctypes.POINTER(ctypes.c_ssize_t)


def _cap_set_err_check(
    result: ctypes.c_int,
    func: Any,
    args: Tuple[Any, ...],
) -> ctypes.c_int:
    if result != 0:
        errno = ctypes.get_errno()
        raise OSError(errno, os.strerror(errno))
    return result


def _cap_get_proc_err_check(
    result: cap_t,
    func: Any,
    args: Tuple[cap_t, cap_flag_t, ctypes.c_int, cap_value_ptr, cap_flag_value_t],
) -> cap_t:
    if not result:
        errno = ctypes.get_errno()
        raise OSError(errno, os.strerror(errno))
    return result


def _cap_get_flag_errcheck(
    result: ctypes.c_int, func: Any, args: Tuple[cap_t, cap_value_t, cap_flag_t, cap_flag_value_ptr]
) -> ctypes.c_int:
    if result != 0:
        errno = ctypes.get_errno()
        raise OSError(errno, os.strerror(errno))
    return result


@cache
def _get_libcap() -> Optional[ctypes.CDLL]:
    if not sys.platform.startswith("linux"):
        raise OSError(f"libcap is only available on Linux, but found platform: {sys.platform}")

    libcap_path = find_library("cap")
    if libcap_path is None:
        LOG.info(
            "Unable to locate libcap. Session action cancelation signals will be sent using sudo"
        )
        return None

    libcap = ctypes.CDLL(libcap_path, use_errno=True)

    # https://man7.org/linux/man-pages/man3/cap_set_proc.3.html
    libcap.cap_set_proc.restype = ctypes.c_int
    libcap.cap_set_proc.argtypes = [
        cap_t,
    ]
    libcap.cap_set_proc.errcheck = _cap_set_err_check  # type: ignore

    # https://man7.org/linux/man-pages/man3/cap_get_proc.3.html
    libcap.cap_get_proc.restype = cap_t
    libcap.cap_get_proc.argtypes = []
    libcap.cap_get_proc.errcheck = _cap_get_proc_err_check  # type: ignore

    # https://man7.org/linux/man-pages/man3/cap_set_flag.3.html
    libcap.cap_set_flag.restype = ctypes.c_int
    libcap.cap_set_flag.argtypes = [
        cap_t,
        cap_flag_t,
        ctypes.c_int,
        cap_value_ptr,
        cap_flag_value_t,
    ]

    # https://man7.org/linux/man-pages/man3/cap_get_flag.3.html
    libcap.cap_get_flag.restype = ctypes.c_int
    libcap.cap_get_flag.argtypes = (
        cap_t,
        cap_value_t,
        cap_flag_t,
        cap_flag_value_ptr,
    )
    libcap.cap_get_flag.errcheck = _cap_get_flag_errcheck  # type: ignore

    return libcap


def _has_capability(
    *,
    libcap: ctypes.CDLL,
    caps: cap_t,
    capability: int,
    capability_set_type: CapabilitySetType,
) -> bool:
    flag_value = cap_flag_value_t()
    libcap.cap_get_flag(caps, capability, capability_set_type.value, ctypes.byref(flag_value))
    return flag_value.value == CAP_SET


@contextmanager
def try_use_cap_kill() -> Generator[bool, None, None]:
    """
    A context-manager that attempts to leverage the CAP_KILL Linux capability.

    If CAP_KILL is in the current thread's effective set, this context-manager takes no action and
    yields True.

    If CAP_KILL is not in the effective set but is in the permitted set, the context-manager:
        1.  adds CAP_KILL to the effective set before entering the context-manager
        2.  yields True
        3.  clears CAP_KILL from the effective set when exiting the context-manager

    Otherwise, the context-manager does nothing and yields False

    Returns:
        A context manager that yields a bool. See above for details.
    """
    if not sys.platform.startswith("linux"):
        raise OSError(f"Only Linux is supported, but platform is {sys.platform}")

    libcap = _get_libcap()
    # If libcap is not found, we yield False indicating we are not aware of having CAP_KILL
    if not libcap:
        yield False
        return

    caps = libcap.cap_get_proc()

    if _has_capability(
        libcap=libcap,
        caps=caps,
        capability=CAP_KILL,
        capability_set_type=CapabilitySetType.EFFECTIVE,
    ):
        LOG.debug("CAP_KILL is in the thread's effective set")
        # CAP_KILL is already in the effective set
        yield True
    elif _has_capability(
        libcap=libcap,
        caps=caps,
        capability=CAP_KILL,
        capability_set_type=CapabilitySetType.PERMITTED,
    ):
        # CAP_KILL is in the permitted set. We will temporarily add it to the effective set
        LOG.debug("CAP_KILL is in the thread's permitted set. Temporarily adding to effective set")
        cap_value_arr_t = cap_value_t * 1
        cap_value_arr = cap_value_arr_t()
        cap_value_arr[0] = CAP_KILL
        libcap.cap_set_flag(
            caps,
            CAP_EFFECTIVE,
            1,
            cap_value_arr,
            CAP_SET,
        )
        libcap.cap_set_proc(caps)
        try:
            yield True
        finally:
            # Clear CAP_KILL from the effective set
            LOG.debug("Clearing CAP_KILL from the thread's effective set")
            libcap.cap_set_flag(
                caps,
                CAP_EFFECTIVE,
                1,
                cap_value_arr,
                CAP_CLEAR,
            )
            libcap.cap_set_proc(caps)
    else:
        yield False


def main() -> None:
    """A developer debugging entrypoint for testing the try_use_cap_kill() behaviour"""
    import logging

    logging.basicConfig(level=logging.DEBUG)
    logging.getLogger("openjd.sessions").setLevel(logging.DEBUG)

    with try_use_cap_kill() as has_cap_kill:
        LOG.info("Has CAP_KILL: %s", has_cap_kill)


if __name__ == "__main__":
    main()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

from enum import Enum

from openjd.model.v2023_09 import Action as Action_2023_09
from openjd.model.v2023_09 import EmbeddedFiles as EmbeddedFiles_2023_09
from openjd.model.v2023_09 import EmbeddedFileText as EmbeddedFileText_2023_09
from openjd.model.v2023_09 import Environment as Environment_2023_09
from openjd.model.v2023_09 import EnvironmentScript as EnvironmentScript_2023_09
from openjd.model.v2023_09 import StepScript as StepScript_2023_09

# ---- Types to export

EnvironmentIdentifier = str

# Turn this into a Union as new schemas are added.
StepScriptModel = StepScript_2023_09

# Turn this into a Union as new schemas are added.
EnvironmentModel = Environment_2023_09
EnvironmentScriptModel = EnvironmentScript_2023_09


class ActionState(str, Enum):
    RUNNING = "running"
    """The action is actively running."""

    CANCELED = "canceled"
    """The action has been canceled and is no longer running."""

    TIMEOUT = "timeout"
    """The action has been canceled due to reaching its runtime limit."""

    FAILED = "failed"
    """The action is no longer running, and exited with a non-zero
    return code."""

    SUCCESS = "success"
    """The action is no longer running, and exited with a zero
    return code."""


# --- Internal types

# dev note: WHen adding support for new schemas, or when an existing schema adds a
# new kind of embedded file, then make these a Union of all of the relevant types.
EmbeddedFileType = EmbeddedFileText_2023_09
EmbeddedFilesListType = EmbeddedFiles_2023_09

# Turn this into a Union as new schemas are added.
ActionModel = Action_2023_09



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

from datetime import timedelta
from ._logging import LoggerAdapter
from pathlib import Path
from typing import Callable, Optional

from openjd.model import SymbolTable
from openjd.model.v2023_09 import CancelationMode as CancelationMode_2023_09
from openjd.model.v2023_09 import StepScript as StepScript_2023_09
from openjd.model.v2023_09 import (
    CancelationMethodNotifyThenTerminate as CancelationMethodNotifyThenTerminate_2023_09,
)
from ._embedded_files import EmbeddedFilesScope
from ._logging import log_subsection_banner
from ._runner_base import (
    CancelMethod,
    NotifyCancelMethod,
    ScriptRunnerBase,
    ScriptRunnerState,
    TerminateCancelMethod,
)
from ._session_user import SessionUser
from ._types import ActionState, StepScriptModel

__all__ = ("StepScriptRunner",)


class StepScriptRunner(ScriptRunnerBase):
    """Use this to run actions from a Step Script."""

    _script: StepScriptModel
    """The step script that we're running.
    """

    _symtab: SymbolTable
    """Treat this as immutable.
    A SymbolTable containing values for all defined variables in the Step
    Script's scope (exluding any symbols defined within the Step Script itself).
    """

    _session_files_directory: Path
    """The location in the filesystem where embedded files will be materialized.
    """

    def __init__(
        self,
        *,
        logger: LoggerAdapter,
        user: Optional[SessionUser] = None,
        # environment for the subprocess that is run
        os_env_vars: Optional[dict[str, Optional[str]]] = None,
        # The working directory of the session
        session_working_directory: Path,
        # `cwd` for the subprocess that's run
        startup_directory: Optional[Path] = None,
        # Callback to invoke when a running action exits
        callback: Optional[Callable[[ActionState], None]] = None,
        script: StepScriptModel,
        symtab: SymbolTable,
        # Directory within which files/attachments should be materialized
        session_files_directory: Path,
    ):
        """
        Arguments (from base class):
            logger (LoggerAdapter): The logger to which all messages should be sent from this and the
                subprocess.
            os_env_vars (dict[str, str]): Environment variables and their values to inject into the
                running subprocess.
            session_working_directory (Path): The temporary directory in which the Session is running.
            user (Optional[SessionUser]): The user to run the subprocess as, if given. Defaults to the
                current user.
            startup_directory (Optional[Path]): cwd to set for the subprocess, if it's possible to set it.
            callback (Optional[Callable[[ActionState], None]]): Callback to invoke when the running
                subprocess has started,  exited, or failed to start. Defaults to None.
        Arguments (unique to this class):
            script (StepScriptModel): The Step Script model that we're going to be running.
            symtab (SymbolTable): A SymbolTable containing values for all defined variables in the Step
                Script's scope (exluding any symbols defined within the Step Script itself).
            session_files_directory (Path): The location in the filesystem where embedded files will
                be materialized.
        """
        super().__init__(
            logger=logger,
            user=user,
            os_env_vars=os_env_vars,
            session_working_directory=session_working_directory,
            startup_directory=startup_directory,
            callback=callback,
        )
        self._script = script
        self._symtab = symtab
        self._session_files_directory = session_files_directory

        if not isinstance(self._script, StepScript_2023_09):
            raise NotImplementedError("Unknown model type")

    def run(self) -> None:
        """Run the Step Script's onRun Action."""
        if self.state != ScriptRunnerState.READY:
            raise RuntimeError("This cannot be used to run a second subprocess.")

        log_subsection_banner(self._logger, "Phase: Setup")

        # For the type checker.
        assert isinstance(self._script, StepScript_2023_09)
        # Write any embedded files to disk
        if self._script.embeddedFiles is not None:
            symtab = SymbolTable(source=self._symtab)
            self._materialize_files(
                EmbeddedFilesScope.STEP,
                self._script.embeddedFiles,
                self._session_files_directory,
                symtab,
            )
            if self.state == ScriptRunnerState.FAILED:
                return
        else:
            symtab = self._symtab

        # Construct the command by evalutating the format strings in the command
        self._run_action(self._script.actions.onRun, symtab)

    def cancel(
        self, *, time_limit: Optional[timedelta] = None, mark_action_failed: bool = False
    ) -> None:
        # For the type checker.
        assert isinstance(self._script, StepScript_2023_09)

        method: CancelMethod
        if (
            self._script.actions.onRun.cancelation is None
            or self._script.actions.onRun.cancelation.mode == CancelationMode_2023_09.TERMINATE
        ):
            # Note: Default cancelation for a 2023-09 Step Script is Terminate
            method = TerminateCancelMethod()
        else:
            model_cancel_method = self._script.actions.onRun.cancelation
            # For the type checker
            assert isinstance(model_cancel_method, CancelationMethodNotifyThenTerminate_2023_09)
            if model_cancel_method.notifyPeriodInSeconds is None:
                # Default grace period is 120s for a 2023-09 Step Script's notify cancel
                method = NotifyCancelMethod(terminate_delay=timedelta(seconds=120))
            else:
                method = NotifyCancelMethod(
                    terminate_delay=timedelta(seconds=model_cancel_method.notifyPeriodInSeconds)
                )

        # Note: If the given time_limit is less than that in the method, then the time_limit will be what's used.
        self._cancel(method, time_limit, mark_action_failed)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import glob
import os
import sys
import time
from subprocess import Popen, DEVNULL, PIPE, STDOUT, run
from typing import Optional

from .._logging import LoggerAdapter, LogContent, LogExtraInfo
from .._os_checker import is_posix, is_linux


class FindSignalTargetError(Exception):
    """Exception when unable to detect the signal target"""

    pass


def find_sudo_child_process_group_id(
    *,
    logger: LoggerAdapter,
    sudo_process: Popen,
    timeout_seconds: float = 1,
) -> Optional[int]:
    # Hint to mypy to not raise module attribute errors (e.g. missing os.getpgid)
    if sys.platform == "win32":
        raise NotImplementedError("This method is for POSIX hosts only")
    if not is_posix():
        raise NotImplementedError(f"Only POSIX supported, but running on {sys.platform}")
    if timeout_seconds <= 0:
        raise ValueError(f"Expected positive value for timeout_seconds but got {timeout_seconds}")

    # For cross-user support, we use sudo which creates an intermediate process:
    #
    #    openjd-process
    #      |
    #      +-- sudo
    #            |
    #            +-- subprocess
    #
    # Sudo forwards signals that it is able to handle, but in the case of SIGKILL sudo cannot
    # handle the signal and the kernel will kill it leaving the child orphaned. We need to
    # send SIGKILL signals to the subprocess of sudo
    start = time.monotonic()
    now = start
    sudo_pgid = os.getpgid(sudo_process.pid)

    # Repeatedly scan for child processes
    #
    # This is put in a retry loop, because it takes a non-zero amount of time before sudo and
    # the kernel finish creating the subprocess. We cap this because the process may exit
    # quickly and we may never find the child process.
    sudo_child_pid: Optional[int] = None
    sudo_child_pgid: Optional[int] = None
    try:
        while now - start < timeout_seconds:
            if not sudo_child_pid:
                if is_linux():
                    sudo_child_pid = find_sudo_child_process_id_procfs(
                        sudo_pid=sudo_process.pid,
                        logger=logger,
                    )
                else:
                    sudo_child_pid = find_child_process_id_pgrep(
                        sudo_pid=sudo_process.pid,
                    )

            if sudo_child_pid:
                try:
                    sudo_child_pgid = os.getpgid(sudo_child_pid)
                except ProcessLookupError:
                    # If the process has exited, we short-circuit
                    return None
                # sudo first forks, then creates a new process group. There is a race condition
                # where the process group ID we observe has not yet changed. If the PGID detected
                # matches the PGID of sudo, then we retry again in the loop
                if sudo_child_pgid == sudo_pgid:
                    sudo_child_pgid = None
                else:
                    break

            # If we did not find any child processes yet, sleep for some time and retry
            time.sleep(min(0.05, timeout_seconds - (now - start)))
            now = time.monotonic()
        if not sudo_child_pid or not sudo_child_pgid:
            raise FindSignalTargetError("unable to detect subprocess before timeout")
    except FindSignalTargetError as e:
        logger.warning(
            f"Unable to determine signal target: {e}",
            extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
        )

    if sudo_child_pgid:
        logger.debug(
            f"Signal target PGID = {sudo_child_pgid}",
            extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
        )

    return sudo_child_pgid


def find_sudo_child_process_id_procfs(
    *,
    logger: LoggerAdapter,
    sudo_pid: int,
) -> Optional[int]:
    # Look for the child process of sudo using procfs. See
    # https://docs.kernel.org/filesystems/proc.html#proc-pid-task-tid-children-information-about-task-children

    child_pids: set[int] = set()
    for task_children_path in glob.glob(f"/proc/{sudo_pid}/task/**/children"):
        with open(task_children_path, "r") as f:
            child_pids.update(int(pid_str) for pid_str in f.read().split())

    # If we found exactly one child, we return it
    if len(child_pids) == 1:

        child_pid = child_pids.pop()

        logger.debug(
            f"Session action process (sudo child) PID is {child_pid}",
            extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
        )
        return child_pid
    # If we found multiple child processes, this violates our assumptions about how sudo
    # works. We will fall-back to using pkill for signalling the process
    elif len(child_pids) > 1:
        raise FindSignalTargetError(
            f"Expected single child processes of sudo, but found {child_pids}"
        )
    return None


def find_child_process_id_pgrep(
    *,
    sudo_pid: int,
) -> Optional[int]:
    pgrep_result = run(
        ["pgrep", "-P", str(sudo_pid)],
        stdout=PIPE,
        stderr=STDOUT,
        stdin=DEVNULL,
        text=True,
    )
    if pgrep_result.returncode != 0:
        raise FindSignalTargetError("Unable to query child processes of sudo process")
    results = pgrep_result.stdout.splitlines()
    if len(results) > 1:
        raise FindSignalTargetError(f"Expected a single child process of sudo, but found {results}")
    elif len(results) == 0:
        return None
    sudo_subproc_pid = int(results[0])
    return sudo_subproc_pid



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import json
import os
import stat
import shlex
from abc import ABC, abstractmethod
from concurrent.futures import Future, ThreadPoolExecutor
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from enum import Enum
from pathlib import Path
from threading import Lock, Timer
from typing import Callable, Optional, Sequence, Type, cast
from types import TracebackType
from tempfile import mkstemp

from openjd.model import SymbolTable
from openjd.model import FormatStringError
from openjd.model.v2023_09 import Action as Action_2023_09
from ._embedded_files import EmbeddedFiles, EmbeddedFilesScope, write_file_for_user
from ._logging import log_subsection_banner, LoggerAdapter, LogContent, LogExtraInfo
from ._os_checker import is_posix
from ._session_user import SessionUser
from ._subprocess import LoggingSubprocess
from ._types import ActionModel, ActionState, EmbeddedFilesListType
from ._win32._locate_executable import locate_windows_executable

__all__ = (
    "ScriptRunnerState",
    "CancelMethod",
    "TerminateCancelMethod",
    "NotifyCancelMethod",
    "ScriptRunnerBase",
)


class ScriptRunnerState(str, Enum):
    """State of a ScriptRunner."""

    READY = "ready"
    """Runner is not currently running anything, and can run an Action.
    """

    RUNNING = "running"
    """Runner is actively running an Action.
    """

    CANCELING = "canceling"
    """Runner is actively in the act of canceling a running Action.
    """

    CANCELED = "canceled"
    """The Action that was run by the runner was canceled.
    """

    TIMEOUT = "timeout"
    """The action has been canceled due to reaching its runtime limit."""

    FAILED = "failed"
    """Runner is done running the subprocess, and the subprocess failed.
    """

    SUCCESS = "success"
    """Runner is done running the subprocess, and the subprocess returned success.
    """


class CancelMethod:
    pass


@dataclass(frozen=True)
class TerminateCancelMethod(CancelMethod):
    """Immediately terminate the running subprocess via SIGKILL"""

    pass


TIME_FORMAT_STR = "%Y-%m-%dT%H:%M:%SZ"


@dataclass(frozen=True)
class NotifyCancelMethod(CancelMethod):
    """Cancelation via "notify then terminate". First send a SIGTERM,
    then wait for a delay, then send a SIGKILL if the process is still running.
    """

    terminate_delay: timedelta
    """Amount of time after a SIGTERM to wait to do the SIGKILL"""


class ScriptRunnerBase(ABC):
    """Base class for a runnable Environment or Step Script.
    Responsible for running a *single* Action, and optionally canceling it.
    """

    _logger: LoggerAdapter
    """The logger to which all messages should be sent from this and the subprocess.
    """

    _user: Optional[SessionUser]
    """The user to run the subprocess as, if given.
    Else the subprocess is run as this process' user.
    """

    _os_env_vars: Optional[dict[str, Optional[str]]]
    """OS Environment variables and their values to inject into the running subprocess.
    """

    _session_working_directory: Path
    """The temporary directory in which the Session is running.
    """

    _startup_directory: Optional[Path]
    """cwd to set for the subprocess, if it's possible to set it.
    """

    _callback: Optional[Callable[[ActionState], None]]
    """Callback to invoke when the running subprocess has exited (or failed to start).
    """

    _process: Optional[LoggingSubprocess]
    """The subprocess that this runner is running, or has most recently run.
    """

    _run_future: Optional[Future]
    """The future within which the current Action for this runner is running.
    Will be None if no Action is running.
    """

    _cancel_gracetime_timer: Optional[Timer]
    """If not None, then this is a timer that is counting down the grace time
    for a NOTIFY_THEN_TERMINATE cancelation.
    self._on_notify_period_end() will be called when the timer expires.
    """

    _cancel_gracetime_end: Optional[datetime]
    """The time at which the gracetime of a NOTIFY_THEN_TERMINATE cancelation's
    graceperiod will expire.
    """

    _canceled: bool
    """True iff the subprocess was canceled.
    """

    _executable_not_found: bool
    """True only on Windows, and only if the executable command that we were given
    could not be found before even trying to run the subprocess.
    """

    _notify_canceled_action_as_failed: bool
    """True iff the subprocess was canceled but action needs to be notified as FAILED.
    """

    _runtime_limit: Optional[Timer]
    """The Timer that will fire when the currently running Action has exhausted
    its runtime limit.
    Will be None if either no Action is running or the running Action has no time
    limit.
    """

    _runtime_limit_reached: bool
    """True if and only if the Action was terminated due to reaching its runtime limit."""

    _pool: ThreadPoolExecutor
    """Pool in which to run futures for this runner.
    """

    _lock: Lock
    """A lock that must be obtained prior to mutating/creating the subprocess
    running state of this runner.
    """

    _state_override: Optional[ScriptRunnerState]
    """An override for subclasses to use to indicate that the runner is in a specific state.
    e.g. We failed to write embedded files before even trying to run the action.
    """

    def __init__(
        self,
        *,
        logger: LoggerAdapter,
        user: Optional[SessionUser] = None,
        # environment for the subprocess that is run
        os_env_vars: Optional[dict[str, Optional[str]]] = None,
        # The working directory of the session
        session_working_directory: Path,
        # `cwd` for the subprocess that's run
        startup_directory: Optional[Path] = None,
        # Callback to invoke when a running action exits
        callback: Optional[Callable[[ActionState], None]] = None,
    ):
        """
        Arguments:
            logger (Logger): The logger to which all messages should be sent from this and the
                subprocess.
            os_env_vars (dict[str, str]): Environment variables and their values to inject into the
                running subprocess.
            session_working_directory (Path): The temporary directory in which the Session is running.
            user (Optional[SessionUser]): The user to run the subprocess as, if given. Defaults to the
                current user.
            startup_directory (Optional[Path]): cwd to set for the subprocess, if it's possible to set it.
            callback (Optional[Callable[[ActionState], None]]): Callback to invoke when the running
                subprocess has started,  exited, or failed to start. Defaults to None.
        """

        self._logger = logger
        self._user = user
        self._os_env_vars = os_env_vars
        self._session_working_directory = session_working_directory
        self._startup_directory = startup_directory
        self._callback = callback

        self._process = None
        self._run_future = None
        self._cancel_gracetime_timer = None
        self._cancel_gracetime_end = None
        self._canceled = False
        self._notify_canceled_action_as_failed = False
        self._runtime_limit = None
        self._runtime_limit_reached = False
        self._executable_not_found = False
        self._lock = Lock()
        # Will run at most the run futures
        self._pool = ThreadPoolExecutor(max_workers=1)
        self._state_override = None

    # Context manager for use in our tests
    def __enter__(self) -> "ScriptRunnerBase":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.shutdown()

    def shutdown(self) -> None:
        """Performs a clean shutdown on the runner. This shutsdown the internal
        ThreadPoolExectutor.
        """
        self._pool.shutdown()

    @abstractmethod
    def cancel(
        self, *, time_limit: Optional[timedelta] = None, mark_action_failed: bool = False
    ) -> None:  # pragma: nocover
        """Cancel the runner's running Action according to whatever method is dictated
        by the specific script being run.

        Arguments:
            time_limit (Optional[timedelta]): If provided, then the cancel must be
                completed within the given number of seconds. This is for urgent
                cancels (e.g. in response to the controlling process getting a SIGTERM).
                Note: a value of 0 turns a notify-then-terminate cancel into a terminate
        """
        raise NotImplementedError("Derived class must implement this.")

    @property
    def state(self) -> ScriptRunnerState:
        """Get the state of this runner."""
        if self._state_override is not None:
            return self._state_override
        if self._process is None:
            return ScriptRunnerState.READY
        # Check on the state of the future for done/canceled first
        # If the future is done, then we have a terminal state.
        assert self._run_future is not None  # For the type checker
        if self._run_future.done():
            if self._canceled and self._notify_canceled_action_as_failed:
                return ScriptRunnerState.FAILED
            if self._canceled and self._runtime_limit_reached:
                return ScriptRunnerState.TIMEOUT
            elif self._canceled:
                return ScriptRunnerState.CANCELED
            elif self._process.failed_to_start or self._process.exit_code != 0:
                return ScriptRunnerState.FAILED
            else:
                return ScriptRunnerState.SUCCESS
        # Future's still running, so we're CANCELING if we've been canceled
        # otherwise we're RUNNING.
        if self._canceled:
            return ScriptRunnerState.CANCELING
        # If the future's not done, then we're still running.
        return ScriptRunnerState.RUNNING

    @property
    def runtime_limit_reached(self) -> bool:
        return self._runtime_limit_reached

    @property
    def exit_code(self) -> Optional[int]:
        """Note: It *is* possible to fail without an exit code."""
        if self._process is not None:
            return self._process.exit_code
        return None

    def _run(self, args: Sequence[str], time_limit: Optional[timedelta] = None) -> None:
        with self._lock:
            if self.state != ScriptRunnerState.READY:
                raise RuntimeError("This cannot be used to run a second subprocess.")
            if is_posix():
                script = self._generate_command_shell_script(args)
                filehandle, filename = mkstemp(
                    dir=self._session_working_directory, suffix=".sh", text=True
                )
                os.close(filehandle)
                # Create the shell script, and make it runnable by the owner.
                # If user is defined, then this will make it owned by that user's group.
                write_file_for_user(
                    Path(filename),
                    script,
                    user=self._user,
                    additional_permissions=stat.S_IXUSR | stat.S_IXGRP,
                )
                self._logger.debug(
                    f"Wrote the following script to {filename}:\n{script}",
                    extra=LogExtraInfo(
                        openjd_log_content=LogContent.FILE_PATH | LogContent.FILE_CONTENTS
                    ),
                )
            else:
                try:
                    args = locate_windows_executable(
                        args, self._user, self._os_env_vars, str(self._session_working_directory)
                    )
                except RuntimeError as e:
                    # Make use of the action filter to surface the failure reason to
                    # the customer.
                    self._logger.info(
                        f"openjd_fail: {str(e)}",
                        extra=LogExtraInfo(openjd_log_content=LogContent.EXCEPTION_INFO),
                    )
                    self._state_override = ScriptRunnerState.FAILED
                    # We haven't started the future yet that runs the process,
                    # but the Session still needs to know that the action is over.
                    if self._callback is not None:
                        self._callback(ActionState.FAILED)
                    return

            subprocess_args = [filename] if is_posix() else args
            self._process = LoggingSubprocess(
                logger=self._logger,
                args=subprocess_args,
                user=self._user,
                os_env_vars=self._os_env_vars,
                working_dir=str(self._session_working_directory),
            )

            if time_limit:
                self._runtime_limit = Timer(time_limit.total_seconds(), self._on_timelimit)
                self._runtime_limit.start()

            log_subsection_banner(self._logger, "Phase: Running action")
            self._run_future = self._pool.submit(self._process.run)
        # Intentionally leave the lock section. If the process was *really* fast,
        # then it's possible for the future to have finished before we get to add
        # the done-callback. That results in the done-callback being called from
        # *this* thread.
        self._run_future.add_done_callback(self._on_process_exit)

        # Block until the subprocess actually starts.
        # This will prevent race conditions where the user starts up the Action,
        # and then is erroneously told that the Action is done because the future
        # for _process.run hasn't actually gotten far enough to start the subprocess
        # before we check self.state
        self._process.wait_until_started()

        if self.state == ScriptRunnerState.RUNNING and self._callback is not None:
            # Let the caller know that the process is running.
            self._callback(ActionState.RUNNING)

    def _generate_command_shell_script(self, args: Sequence[str]) -> str:
        """Generate a shell script for running a command given by the args."""
        script = list[str]()
        script.append("#!/bin/sh\n")
        if self._os_env_vars:
            for name, value in self._os_env_vars.items():
                if value is None:
                    script.append(f"unset {name}")
                else:
                    script.append(f"export {name}={shlex.quote(value)}")
        if self._startup_directory is not None:
            # Note: Single quotes around the path as it may have spaces, and we don't want to
            # process any shell commands in the path.
            script.append(f"cd '{self._startup_directory}'")
        script.append("exec " + shlex.join(args))
        return "\n".join(script)

    def _materialize_files(
        self,
        scope: EmbeddedFilesScope,
        files: EmbeddedFilesListType,
        dest_directory: Path,
        symtab: SymbolTable,
    ) -> None:
        """Helper for derived classes that wraps all of the logic around
        materializing embedded files to disk.

        """
        file_writer = EmbeddedFiles(
            logger=self._logger,
            scope=scope,
            session_files_directory=dest_directory,
            user=self._user,
        )
        try:
            file_writer.materialize(files, symtab)
        except RuntimeError as exc:
            # Had a problem writing at least one file to disk.
            # Surface the error.
            # Make use of the action filter to surface the failure reason to
            # the customer.
            self._logger.info(
                f"openjd_fail: {str(exc)}",
                extra=LogExtraInfo(openjd_log_content=LogContent.EXCEPTION_INFO),
            )
            self._state_override = ScriptRunnerState.FAILED
            # We haven't started the future yet that runs the process,
            # but the Session still needs to know that the action is over.
            if self._callback is not None:
                self._callback(ActionState.FAILED)

    def _run_action(
        self,
        action: ActionModel,
        symtab: SymbolTable,
        *,
        default_timeout: Optional[timedelta] = None,
    ) -> None:
        """Helper for derived classes to run a specific Action.

        Args:
            action (ActionModel): The action model to be executed. Must be an
                instance of Action_2023_09.
            symtab (SymbolTable): Symbol table used for resolving command and
                arguments.
            default_timeout (Optional[timedelta], optional): Default timeout duration
                for the action if no timeout is specified in the action. The default behaviour if
                None is passed will allow the action to run indefinitely until it completes.
        """
        assert isinstance(action, Action_2023_09)
        try:
            command = [action.command.resolve(symtab=symtab)]
            if action.args is not None:
                command.extend(s.resolve(symtab=symtab) for s in action.args)
        except FormatStringError as exc:
            # Extremely unlikely since a JobTemplate needs to have passed
            # validation before we could be running it, but just to be safe.
            self._logger.info(
                f"openjd_fail: {str(exc)}",
                extra=LogExtraInfo(openjd_log_content=LogContent.EXCEPTION_INFO),
            )
            self._state_override = ScriptRunnerState.FAILED
            # We haven't started the future yet that runs the process,
            # but the Session still needs to know that the action is over.
            if self._callback is not None:
                self._callback(ActionState.FAILED)
        else:
            time_limit: Optional[timedelta] = default_timeout
            if action.timeout:
                time_limit = timedelta(seconds=action.timeout)
            self._run(command, time_limit)

    def _cancel(
        self,
        method: CancelMethod,
        time_limit: Optional[timedelta] = None,
        mark_action_failed: bool = False,
    ) -> None:
        # For the type checkers
        assert self._process is not None
        # Nothing to do if it's not running.
        if not self._process.is_running:
            return

        with self._lock:
            self._canceled = True
            self._notify_canceled_action_as_failed = mark_action_failed
            now = datetime.now(timezone.utc)
            now_str = now.strftime(TIME_FORMAT_STR)
            if self._cancel_gracetime_timer is not None:
                # This cancel request is a duplicate that may have a different gracetime.
                # We'll recalculate the gracetime
                self._cancel_gracetime_timer.cancel()
                self._cancel_gracetime_timer = None

            if isinstance(method, TerminateCancelMethod):
                self._logger.info(
                    f"Canceling subprocess {str(self._process.pid)} via termination method at {now_str}.",
                    extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
                )
                try:
                    self._process.terminate()
                except OSError as err:  # pragma: nocover
                    # Being paranoid. Won't happen... if we could start the process, then we can send it a signal
                    self._logger.warning(
                        f"Cancelation could not send terminate signal to process {self._process.pid}: {str(err)}",
                        extra=LogExtraInfo(
                            openjd_log_content=LogContent.PROCESS_CONTROL
                            | LogContent.EXCEPTION_INFO
                        ),
                    )
            else:
                self._logger.info(
                    f"Canceling subprocess {str(self._process.pid)} via notify then terminate method at {now_str}.",
                    extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
                )
                method = cast(NotifyCancelMethod, method)

                gracetime = (
                    min(time_limit, method.terminate_delay)
                    if time_limit is not None
                    else method.terminate_delay
                )
                if self._cancel_gracetime_end is not None:
                    # How much time is remaining in the previous cancel?
                    time_remaining = self._cancel_gracetime_end - now
                    # Our gracetime is the minimum of remaining and the new time limit
                    gracetime = min(gracetime, time_remaining)
                self._cancel_gracetime_end = now + gracetime

                # 1) Create the notification file
                #      Note: Notify-then-terminate requires writing a "cancel_info.json" file to
                #      the session working directory. Contents are JSON formatted with contents:
                #      { "NotifyEnd": "<yyyy>-<mm>-<dd>T<hh>:<mm>:<ss>Z" }
                #      where the given time is the time at which the notify period will end (i.e. when
                #      when we'll send the SIGKILL)
                grace_end_time_str = self._cancel_gracetime_end.strftime(TIME_FORMAT_STR)
                notify_end = json.dumps({"NotifyEnd": grace_end_time_str})
                write_file_for_user(
                    self._session_working_directory / "cancel_info.json", notify_end, self._user
                )
                self._logger.info(
                    f"Grace period ends at {grace_end_time_str}",
                    extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
                )

                # 2) Send the notify
                try:
                    self._process.notify()
                except OSError as err:  # pragma: nocover
                    # Being paranoid. Won't happen... if we could start the process, then we can send it a signal
                    self._logger.warning(
                        f"Cancelation could not send notify signal to process {self._process.pid}: {str(err)}",
                        extra=LogExtraInfo(
                            openjd_log_content=LogContent.PROCESS_CONTROL
                            | LogContent.EXCEPTION_INFO
                        ),
                    )

                # 4) Set up the timer to send the terminate signal
                self._cancel_gracetime_timer = Timer(
                    gracetime.total_seconds(), self._on_notify_period_end
                )
                self._cancel_gracetime_timer.start()

    def _on_process_exit(self, future: Future) -> None:
        """This is invoked as a callback when run_future is done."""
        assert self._run_future is not None
        with self._lock:
            if self._runtime_limit is not None:
                self._runtime_limit.cancel()
                self._runtime_limit = None

            if self._cancel_gracetime_timer is not None:
                self._cancel_gracetime_timer.cancel()
                self._cancel_gracetime_timer = None

            if exc := self._run_future.exception():
                self._logger.error(
                    f"Error running subprocess: {str(exc)}",
                    extra=LogExtraInfo(
                        openjd_log_content=LogContent.PROCESS_CONTROL | LogContent.EXCEPTION_INFO
                    ),
                )

            if self._callback is not None:
                self._callback(ActionState(self.state.value))

    def _on_notify_period_end(self) -> None:
        """This is invoked when the grace period in a NOTIFY_THEN_TERMINATE
        cancelation has expired.
        """
        assert self._process is not None
        with self._lock:
            self._cancel_gracetime_timer = None
        self._logger.info(
            "Notify period ended. Terminate at %s",
            datetime.now(timezone.utc).strftime(TIME_FORMAT_STR),
            extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
        )
        try:
            self._process.terminate()
        except OSError as err:  # pragma: nocover
            # Being paranoid. Won't happen... if we could start the process, then we can send it a kill signal
            self._logger.warning(
                f"Cancelation could not send terminate signal to process {self._process.pid}: {str(err)}",
                extra=LogExtraInfo(
                    openjd_log_content=LogContent.PROCESS_CONTROL | LogContent.EXCEPTION_INFO
                ),
            )

    def _on_timelimit(self) -> None:
        """Callback that is invoked when the runtime limit of the running
        process has expired.
        """
        with self._lock:
            self._runtime_limit = None
        self._logger.info(
            "TIMEOUT - Runtime limit reached at %s. Canceling action.",
            datetime.now(timezone.utc).strftime(TIME_FORMAT_STR),
            extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
        )
        self._runtime_limit_reached = True
        self.cancel()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import os
import stat
from ._logging import LoggerAdapter, LogContent, LogExtraInfo
from pathlib import Path
from shutil import chown, rmtree
from tempfile import gettempdir, mkdtemp
from typing import Optional, cast

from ._session_user import PosixSessionUser, SessionUser, WindowsSessionUser
from ._windows_permission_helper import WindowsPermissionHelper
from ._os_checker import is_posix, is_windows

if is_windows():
    from ._win32._helpers import get_process_user  # type: ignore


def custom_gettempdir(logger: Optional[LoggerAdapter] = None) -> str:
    """
    Get a platform-specific temporary directory.

    For Windows systems, this function returns a specific directory path,
    '%PROGRAMDATA%\\Amazon\\'. If this directory does not exist, it will be created.
    For non-Windows systems, it returns the system's default temporary directory.

    Args:
        logger (Optional[LoggerAdapter]): The logger to which all messages should be sent from this and the
                subprocess.

    Returns:
        str: The path to the temporary directory specific to the operating system.
    """
    if is_windows():
        program_data_path = os.getenv("PROGRAMDATA")
        if program_data_path is None:
            program_data_path = r"C:\ProgramData"
            if logger:
                logger.warning(
                    f'Environment variable "PROGRAMDATA" is not set. Creating the session working directories under {program_data_path}',
                    extra=LogExtraInfo(openjd_log_content=LogContent.FILE_PATH),
                )

        temp_dir_parent = os.path.join(program_data_path, "Amazon")
    else:
        temp_dir_parent = gettempdir()

    temp_dir = os.path.join(temp_dir_parent, "OpenJD")
    os.makedirs(temp_dir, exist_ok=True)
    return temp_dir


class TempDir:
    """This class securely creates a temporary directory using the same rules as mkdtemp(),
    but with the option of having the directory owned by a user other than this process' user.

    Notes:
        posix - Only the group of the temp directory is set. The directory owner will be this
            process' uid. This process must be running as root to change the ownership, so we don't
            do it (don't really need to, either, since the use-case for this class is to
            create the Open Job Description Session working directory and that working directory needs to be
            both writable and deletable by this process).
    """

    path: Path
    """Pathname of the created directory.
    """

    def __init__(
        self,
        *,
        dir: Optional[Path] = None,
        prefix: Optional[str] = None,
        user: Optional[SessionUser] = None,
        logger: Optional[LoggerAdapter] = None,
    ):
        """
        Arguments:
            dir (Optional[Path]): The directory in which to create the temp dir.
                Defaults to tempfile.gettempdir().
            prefix (Optional[str]): A prefix to use in the name of the generated temp dir.
                Defaults to "".
            user (Optional[SessionUser]): A group that will own the created directory.
                The group-write bit will be set on the directory if this option is supplied.
                Defaults to this process' effective user/group.
            logger (Optional[LoggerAdapter]): The logger to which all messages should be sent from this and the
                subprocess.

        Raises:
            RuntimeError - If this process cannot create the temporary directory, or change the
                group ownership of the created directory.
        """
        # pre-flight checks
        if user and is_posix() and not isinstance(user, PosixSessionUser):  # pragma: nocover
            raise ValueError("user must be a posix-user. Got %s", type(user))
        elif user and is_windows() and not isinstance(user, WindowsSessionUser):
            raise ValueError("user must be a windows-user. Got %s", type(user))

        if not dir:
            dir = Path(custom_gettempdir(logger))

        dir = dir.resolve()
        try:
            self.path = Path(mkdtemp(dir=dir, prefix=prefix))  # 0o700
        except OSError as err:
            raise RuntimeError(f"Could not create temp directory within {str(dir)}: {str(err)}")

        # Change the owner
        if user:
            if is_posix():
                user = cast(PosixSessionUser, user)
                # Change ownership
                try:
                    chown(self.path, group=user.group)
                except OSError as err:
                    raise RuntimeError(
                        f"Could not change ownership of directory '{str(dir)}' (error: {str(err)}). Please ensure that uid {os.geteuid()} is a member of group {user.group}."  # type: ignore
                    )
                # Update the permissions to include the group after the group is changed
                # Note: Only after changing group for security in case the group-ownership
                # change fails.
                os.chmod(self.path, mode=stat.S_IRWXU | stat.S_IRWXG)
            elif is_windows():
                user = cast(WindowsSessionUser, user)
                try:
                    WindowsPermissionHelper.set_permissions(
                        str(self.path),
                        principals_full_control=[get_process_user()],
                        principals_modify_access=[user.user],
                    )
                except Exception as err:
                    raise RuntimeError(
                        f"Could not change permissions of directory '{str(dir)}' (error: {str(err)})"
                    )

    def cleanup(self) -> None:
        """Deletes the temporary directory and all of its contents.
        Raises:
            RuntimeError - If not all files could be deleted.
        """
        encountered_errors = False
        file_paths: list[str] = []

        def onerror(f, p, e):
            nonlocal encountered_errors
            nonlocal file_paths
            encountered_errors = True
            file_paths.append(str(p))

        rmtree(self.path, onerror=onerror)
        if encountered_errors:
            raise RuntimeError(
                f"Files within temporary directory {str(self.path)} could not be deleted.\n"
                + "\n".join(file_paths)
            )
